"use strict";
var __awaiter = (this && this.__awaiter) || function (thisArg, _arguments, P, generator) {
    function adopt(value) { return value instanceof P ? value : new P(function (resolve) { resolve(value); }); }
    return new (P || (P = Promise))(function (resolve, reject) {
        function fulfilled(value) { try { step(generator.next(value)); } catch (e) { reject(e); } }
        function rejected(value) { try { step(generator["throw"](value)); } catch (e) { reject(e); } }
        function step(result) { result.done ? resolve(result.value) : adopt(result.value).then(fulfilled, rejected); }
        step((generator = generator.apply(thisArg, _arguments || [])).next());
    });
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.getOraclePrice = exports.decodeRecentEvents = exports.findLargestTokenAccountForOwner = exports.getMultipleAccounts = exports.parseTokenAccount = exports.parseTokenAccountData = exports.ACCOUNT_LAYOUT = exports.getUnixTs = exports.promiseUndef = exports.getFilteredProgramAccounts = exports.nativeToUi = exports.uiToNative = exports.getMintDecimals = exports.createAccountInstruction = exports.awaitTransactionSignatureConfirmation = exports.simulateTransaction = exports.getDecimalCount = exports.sleep = exports.zeroKey = void 0;
const web3_js_1 = require("@solana/web3.js");
const bn_js_1 = __importDefault(require("bn.js"));
const token_instructions_1 = require("@project-serum/serum/lib/token-instructions");
const client_1 = require("@pythnetwork/client");
const buffer_layout_1 = require("buffer-layout");
const spl_token_1 = require("@solana/spl-token");
const layout_1 = require("./layout");
const layout_2 = require("@project-serum/serum/lib/layout");
const schema_1 = require("./schema");
exports.zeroKey = new web3_js_1.PublicKey(new Uint8Array(32));
function sleep(ms) {
    return __awaiter(this, void 0, void 0, function* () {
        return new Promise((resolve) => setTimeout(resolve, ms));
    });
}
exports.sleep = sleep;
function getDecimalCount(value) {
    if (!isNaN(value) &&
        Math.floor(value) !== value &&
        value.toString().includes('.'))
        return value.toString().split('.')[1].length || 0;
    if (!isNaN(value) &&
        Math.floor(value) !== value &&
        value.toString().includes('e'))
        return parseInt(value.toString().split('e-')[1] || '0');
    return 0;
}
exports.getDecimalCount = getDecimalCount;
function simulateTransaction(connection, transaction, commitment) {
    return __awaiter(this, void 0, void 0, function* () {
        // @ts-ignore
        transaction.recentBlockhash = yield connection._recentBlockhash(
        // @ts-ignore
        connection._disableBlockhashCaching);
        const signData = transaction.serializeMessage();
        // @ts-ignore
        const wireTransaction = transaction._serialize(signData);
        const encodedTransaction = wireTransaction.toString('base64');
        const config = { encoding: 'base64', commitment };
        const args = [encodedTransaction, config];
        // @ts-ignore
        const res = yield connection._rpcRequest('simulateTransaction', args);
        if (res.error) {
            throw new Error('failed to simulate transaction: ' + res.error.message);
        }
        return res.result;
    });
}
exports.simulateTransaction = simulateTransaction;
function awaitTransactionSignatureConfirmation(txid, timeout, connection, confirmLevel) {
    return __awaiter(this, void 0, void 0, function* () {
        let done = false;
        const confirmLevels = ['finalized'];
        if (confirmLevel === 'confirmed') {
            confirmLevels.push('confirmed');
        }
        else if (confirmLevel === 'processed') {
            confirmLevels.push('confirmed');
            confirmLevels.push('processed');
        }
        const result = yield new Promise((resolve, reject) => {
            (() => __awaiter(this, void 0, void 0, function* () {
                setTimeout(() => {
                    if (done) {
                        return;
                    }
                    done = true;
                    console.log('Timed out for txid', txid);
                    reject({ timeout: true });
                }, timeout);
                try {
                    connection.onSignature(txid, (result) => {
                        // console.log('WS confirmed', txid, result);
                        done = true;
                        if (result.err) {
                            reject(result.err);
                        }
                        else {
                            resolve(result);
                        }
                    }, 'singleGossip');
                    // console.log('Set up WS connection', txid);
                }
                catch (e) {
                    done = true;
                    console.log('WS error in setup', txid, e);
                }
                while (!done) {
                    // eslint-disable-next-line no-loop-func
                    (() => __awaiter(this, void 0, void 0, function* () {
                        try {
                            const signatureStatuses = yield connection.getSignatureStatuses([
                                txid,
                            ]);
                            const result = signatureStatuses && signatureStatuses.value[0];
                            if (!done) {
                                if (!result) {
                                    // console.log('REST null result for', txid, result);
                                }
                                else if (result.err) {
                                    console.log('REST error for', txid, result);
                                    done = true;
                                    reject(result.err);
                                }
                                else if (!(result.confirmations ||
                                    confirmLevels.includes(result.confirmationStatus))) {
                                    console.log('REST not confirmed', txid, result);
                                }
                                else {
                                    console.log('REST confirmed', txid, result);
                                    done = true;
                                    resolve(result);
                                }
                            }
                        }
                        catch (e) {
                            if (!done) {
                                console.log('REST connection error: txid', txid, e);
                            }
                        }
                    }))();
                    yield sleep(300);
                }
            }))();
        });
        done = true;
        return result;
    });
}
exports.awaitTransactionSignatureConfirmation = awaitTransactionSignatureConfirmation;
function createAccountInstruction(connection, payer, space, owner, lamports) {
    return __awaiter(this, void 0, void 0, function* () {
        const account = new web3_js_1.Account();
        const instruction = web3_js_1.SystemProgram.createAccount({
            fromPubkey: payer,
            newAccountPubkey: account.publicKey,
            lamports: lamports
                ? lamports
                : yield connection.getMinimumBalanceForRentExemption(space),
            space,
            programId: owner,
        });
        return { account, instruction };
    });
}
exports.createAccountInstruction = createAccountInstruction;
const MINT_LAYOUT = buffer_layout_1.struct([buffer_layout_1.blob(44), buffer_layout_1.u8('decimals'), buffer_layout_1.blob(37)]);
function getMintDecimals(connection, mint) {
    return __awaiter(this, void 0, void 0, function* () {
        if (mint.equals(token_instructions_1.WRAPPED_SOL_MINT)) {
            return 9;
        }
        const { data } = throwIfNull(yield connection.getAccountInfo(mint), 'mint not found');
        const { decimals } = MINT_LAYOUT.decode(data);
        return decimals;
    });
}
exports.getMintDecimals = getMintDecimals;
function throwIfNull(value, message = 'account not found') {
    if (value === null) {
        throw new Error(message);
    }
    return value;
}
function uiToNative(amount, decimals) {
    return new bn_js_1.default(Math.round(amount * Math.pow(10, decimals)));
}
exports.uiToNative = uiToNative;
function nativeToUi(amount, decimals) {
    return amount / Math.pow(10, decimals);
}
exports.nativeToUi = nativeToUi;
function getFilteredProgramAccounts(connection, programId, filters) {
    return __awaiter(this, void 0, void 0, function* () {
        // @ts-ignore
        const resp = yield connection._rpcRequest('getProgramAccounts', [
            programId.toBase58(),
            {
                commitment: connection.commitment,
                filters,
                encoding: 'base64',
            },
        ]);
        if (resp.error) {
            throw new Error(resp.error.message);
        }
        return resp.result.map(({ pubkey, account: { data, executable, owner, lamports } }) => ({
            publicKey: new web3_js_1.PublicKey(pubkey),
            accountInfo: {
                data: Buffer.from(data[0], 'base64'),
                executable,
                owner: new web3_js_1.PublicKey(owner),
                lamports,
            },
        }));
    });
}
exports.getFilteredProgramAccounts = getFilteredProgramAccounts;
function promiseUndef() {
    return __awaiter(this, void 0, void 0, function* () {
        return undefined;
    });
}
exports.promiseUndef = promiseUndef;
const getUnixTs = () => {
    return new Date().getTime() / 1000;
};
exports.getUnixTs = getUnixTs;
exports.ACCOUNT_LAYOUT = buffer_layout_1.struct([
    buffer_layout_1.blob(32, 'mint'),
    buffer_layout_1.blob(32, 'owner'),
    buffer_layout_1.nu64('amount'),
    buffer_layout_1.blob(93),
]);
function parseTokenAccountData(data) {
    let { mint, owner, amount } = exports.ACCOUNT_LAYOUT.decode(data);
    return {
        mint: new web3_js_1.PublicKey(mint),
        owner: new web3_js_1.PublicKey(owner),
        amount,
    };
}
exports.parseTokenAccountData = parseTokenAccountData;
function parseTokenAccount(data) {
    const decoded = layout_1.AccountLayout.decode(data);
    return {
        mint: decoded.mint,
        owner: decoded.owner,
        amount: decoded.amount,
    };
}
exports.parseTokenAccount = parseTokenAccount;
function getMultipleAccounts(connection, publicKeys, commitment) {
    return __awaiter(this, void 0, void 0, function* () {
        const len = publicKeys.length;
        if (len > 100) {
            const mid = Math.floor(publicKeys.length / 2);
            return Promise.all([
                getMultipleAccounts(connection, publicKeys.slice(0, mid), commitment),
                getMultipleAccounts(connection, publicKeys.slice(mid, len), commitment),
            ]).then((a) => a[0].concat(a[1]));
        }
        const publicKeyStrs = publicKeys.map((pk) => pk.toBase58());
        // load connection commitment as a default
        commitment || (commitment = connection.commitment);
        const args = commitment ? [publicKeyStrs, { commitment }] : [publicKeyStrs];
        // @ts-ignore
        const resp = yield connection._rpcRequest('getMultipleAccounts', args);
        if (resp.error) {
            throw new Error(resp.error.message);
        }
        return resp.result.value.map(({ data, executable, lamports, owner }, i) => ({
            publicKey: publicKeys[i],
            context: resp.result.context,
            accountInfo: {
                data: Buffer.from(data[0], 'base64'),
                executable,
                owner: new web3_js_1.PublicKey(owner),
                lamports,
            },
        }));
    });
}
exports.getMultipleAccounts = getMultipleAccounts;
function findLargestTokenAccountForOwner(connection, owner, mint) {
    return __awaiter(this, void 0, void 0, function* () {
        const response = yield connection.getTokenAccountsByOwner(owner, { mint, programId: spl_token_1.TOKEN_PROGRAM_ID }, connection.commitment);
        let max = -1;
        let maxTokenAccount = null;
        let maxPubkey = null;
        for (const { pubkey, account } of response.value) {
            const tokenAccount = parseTokenAccountData(account.data);
            if (tokenAccount.amount > max) {
                maxTokenAccount = tokenAccount;
                max = tokenAccount.amount;
                maxPubkey = pubkey;
            }
        }
        if (maxPubkey && maxTokenAccount) {
            return { publicKey: maxPubkey, tokenAccount: maxTokenAccount };
        }
        else {
            throw new Error('No accounts for this token');
        }
    });
}
exports.findLargestTokenAccountForOwner = findLargestTokenAccountForOwner;
const EVENT_QUEUE_HEADER = buffer_layout_1.struct([
    buffer_layout_1.blob(5),
    layout_2.accountFlagsLayout('accountFlags'),
    buffer_layout_1.u32('head'),
    layout_2.zeros(4),
    buffer_layout_1.u32('count'),
    layout_2.zeros(4),
    buffer_layout_1.u32('seqNum'),
    layout_2.zeros(4),
]);
const EVENT_FLAGS = buffer_layout_1.bits(buffer_layout_1.u8(), false, 'eventFlags');
EVENT_FLAGS.addBoolean('fill');
EVENT_FLAGS.addBoolean('out');
EVENT_FLAGS.addBoolean('bid');
EVENT_FLAGS.addBoolean('maker');
const EVENT = buffer_layout_1.struct([
    EVENT_FLAGS,
    buffer_layout_1.u8('openOrdersSlot'),
    buffer_layout_1.u8('feeTier'),
    buffer_layout_1.blob(5),
    layout_2.u64('nativeQuantityReleased'),
    layout_2.u64('nativeQuantityPaid'),
    layout_2.u64('nativeFeeOrRebate'),
    layout_2.u128('orderId'),
    layout_2.publicKeyLayout('openOrders'),
    layout_2.u64('clientOrderId'),
]);
function decodeRecentEvents(buffer, lastSeenSeqNum) {
    const header = EVENT_QUEUE_HEADER.decode(buffer);
    const nodes = [];
    if (lastSeenSeqNum !== undefined) {
        const allocLen = Math.floor((buffer.length - EVENT_QUEUE_HEADER.span) / EVENT.span);
        const newEventsCount = header.seqNum - lastSeenSeqNum;
        for (let i = newEventsCount; i > 0; --i) {
            const nodeIndex = (header.head + header.count + allocLen - i) % allocLen;
            const decodedItem = EVENT.decode(buffer, EVENT_QUEUE_HEADER.span + nodeIndex * EVENT.span);
            nodes.push(decodedItem);
        }
    }
    return { header, nodes };
}
exports.decodeRecentEvents = decodeRecentEvents;
const PYTH_MAGIC = Buffer.from([0xa1, 0xb2, 0xc3, 0xd4]);
function getOraclePrice(connection, oracle) {
    return __awaiter(this, void 0, void 0, function* () {
        const info = yield connection.getAccountInfo(oracle);
        if (!info || !info.data.length) {
            throw new Error('account does not exist');
        }
        const pythBase = client_1.parseBaseData(info.data);
        if ((pythBase === null || pythBase === void 0 ? void 0 : pythBase.type) == client_1.AccountType.Price) {
            const price = client_1.parsePriceData(info.data);
            return price.aggregate.price;
        }
        else {
            const agg = schema_1.Aggregator.deserialize(info.data);
            return agg.answer.median.toNumber() / Math.pow(10, agg.config.decimals);
        }
    });
}
exports.getOraclePrice = getOraclePrice;
//# sourceMappingURL=utils.js.map